from django.apps import AppConfig


class SimpleOauth2Config(AppConfig):
    """Configuration for the simple_oauth2 app."""

    default_auto_field = "django.db.models.BigAutoField"
    name = "simple_oauth2"
