# Generated by Django 4.1.13 on 2025-09-23 02:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("simple_oauth2", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="session",
            name="status",
            field=models.CharField(
                choices=[
                    ("pending", "Pending"),
                    ("token_failed", "Token Failed"),
                    ("userinfo_failed", "Userinfo Failed"),
                    ("expired", "Expired"),
                    ("completed", "Completed"),
                ],
                default="pending",
                max_length=20,
            ),
        ),
    ]
